import logging
import time
from typing import List, Optional, Tuple

from adbc_driver_flightsql import dbapi  # type:ignore

from scandal._auth import auth_user

logging.basicConfig(level=logging.INFO)


def connect(org: Optional[str] = None, project: Optional[str] = None, auth_token: Optional[str] = None):
    """Connect to Scandal.

    By default,
        - For interactive users we log in using an OAuth2 device code flow to the user's default organization.
        - For non-interactive users, we pick up creds from a credentials chain.
    """
    if not auth_token:
        auth_token = auth_user()

    # For now, we send the token as the Basic Auth password. This is because the response headers
    # aren't respected when using Bearer auth: https://github.com/apache/arrow-adbc/issues/584
    db_kwargs = {"username": "scandal", "password": auth_token}

    if project:
        # Connect directly to a project
        db_kwargs[_custom_header("x-fulcrum-project")] = project
    if org:
        db_kwargs[_custom_header("x-fulcrum-org")] = org

    return dbapi.connect("grpc://localhost:8443", db_kwargs)


def _custom_header(name: str):
    # Construct the name of a custom header to pass to the Flight SQL server
    return f"adbc.flight.sql.rpc.call_header.{name}"


if __name__ == "__main__":
    import pandas_gbq  # type:ignore
    from google.oauth2.service_account import Credentials  # type:ignore

    sql = "SELECT * from tpch.lineitem"

    credentials = Credentials.from_service_account_file(
        "../../crowbar/crowbar-server/src/test/resources/bigquery.creds.json"
    )
    conn = connect(org="my-org", project="my-source")

    times: List[Tuple[float, float, float]] = []
    for i in range(10):
        start = time.time()
        df = pandas_gbq.read_gbq(sql, credentials=credentials)
        # df = pandas_gbq.read_gbq("loyal-road-382017.tpch.lineitem", credentials=credentials)
        pdt = time.time() - start
        print("PandasBQ", pdt, "seconds")

        # start = time.time()
        # df = pd.read_sql_query(sql, conn)
        # sdt = time.time() - start
        # print("Scandal", sdt, "seconds")

        start = time.time()
        curr = conn.cursor()
        curr.execute(sql)
        table = curr.fetch_arrow_table()
        df = table.to_pandas()
        sat = time.time() - start
        print("Scandal Arrow", sat, "seconds")
        # print(table)

        # times.append((pdt, sdt, sat))

    print(times)



import time
import webbrowser

import jwt
import keyring
import requests

SERVICE = "scandal"
AUTH_DOMAIN = "http://localhost:3000"
CLIENT_ID = "app_3JQW6ThZHnj2VJMlemk6D"
GRANT_TYPE = "urn:ietf:params:oauth:grant-type:device_code"
# Refresh token 5 minutes before it expires
EXPIRY_WINDOW = 5 * 60


def auth_user():
    # _try_delete_password(SERVICE, "access_token")
    # _try_delete_password(SERVICE, "refresh_token")

    access_token = keyring.get_password(SERVICE, "access_token")
    refresh_token = keyring.get_password(SERVICE, "refresh_token")

    if access_token:
        access_token_decoded = jwt.decode(
            access_token, options={"verify_signature": False}
        )

        if access_token_decoded.get("exp") > (time.time() - EXPIRY_WINDOW):
            return access_token

        if refresh_token:
            refresh_resp = requests.post(
                f"{AUTH_DOMAIN}/oauth/token",
                data={
                    "client_id": CLIENT_ID,
                    "grant_type": GRANT_TYPE,
                    "refresh_token": refresh_token,
                },
            )
            if not refresh_resp.ok:
                print("Failed to refresh Scandal token.")
                keyring.delete_password(SERVICE, "access_token")
                keyring.delete_password(SERVICE, "refresh_token")
                return auth_user()

    resp = requests.post(
        f"{AUTH_DOMAIN}/oauth/device",
        data={
            "client_id": CLIENT_ID,
            # TODO(ngates): should this be scandal instead?
            "audience": "https://scandal.fulcrum.so",
            "scope": "read offline_access",
        },
    )
    if not resp.ok:
        raise Exception(f"{resp.status_code}: {resp.text}")
    code_resp = resp.json()

    expiry = time.time() + code_resp["expires_in"]

    print("Enter the following code at this URL: " + code_resp["verification_uri"])
    print(code_resp["user_code"])
    webbrowser.open(code_resp["verification_uri_complete"])

    while time.time() < expiry:
        time.sleep(code_resp["interval"])
        token_resp = requests.post(
            f"{AUTH_DOMAIN}/oauth/token",
            data={
                "client_id": CLIENT_ID,
                "grant_type": GRANT_TYPE,
                "device_code": code_resp["device_code"],
            },
        )

        if token_resp.ok:
            token_resp = token_resp.json()
            access_token = token_resp["access_token"]
            refresh_token = token_resp.get("refresh_token", "refresh")

            keyring.set_password(SERVICE, "access_token", access_token)
            keyring.set_password(SERVICE, "refresh_token", refresh_token)

            return access_token

        if token_resp.json()["error"] != "authorization_pending":
            token_resp.raise_for_status()

    raise Exception("Timed out waiting for authorization. Please try again")


def _try_delete_password(service: str, username: str):
    try:
        keyring.delete_password(service, username)
    except Exception:
        pass
